
#ifndef _TEST_UNION_DIFFERENT_ITERATOR
#define _TEST_UNION_DIFFERENT_ITERATOR

#include <assert.h>
#include <vector>
#include <iostream>
#include <algorithm>
#include <utility>

#include "Test.hpp"
#include "Test_helper.hpp"
#include "union_different_iterator.hpp"

// concepts required: FI1 ed FI2 devono dereferenziare un tipo per cui � definito
// operator<<() su un ostream
template <typename FI1, typename FI2,  template <typename> class WR_A, template <typename> class WR_B>
void print_union_different( const union_different_iterator<FI1, FI2, WR_A, WR_B>& ui,
						    const union_different_iterator<FI1, FI2, WR_A, WR_B>& ui_end )
{
    using namespace std;
	typedef typename union_different_iterator<FI1, FI2, WR_A, WR_B>::value_type1 value_type1;
	typedef typename union_different_iterator<FI1, FI2, WR_A, WR_B>::value_type2 value_type2;

	union_different_iterator<FI1, FI2, WR_A, WR_B> local_ui = ui;

	cout << "Union: ";
	for ( ; local_ui != ui_end; ++local_ui )
	{
		if ( local_ui.template isTypeOf<value_type1>() )
			cout << boost::get<WR_A<value_type1> >( *local_ui )._data << " ";
		else
			cout << boost::get<WR_B<value_type2> >( *local_ui )._data << " ";
	}
	cout << endl;
}


template <typename FI1, typename FI2, template <typename> class WR_A, template <typename> class WR_B>
void print_union_different( const std::pair< union_different_iterator<FI1, FI2, WR_A, WR_B>,
				                             union_different_iterator<FI1, FI2, WR_A, WR_B> > union_range )
{
	print_union_different( union_range.first, union_range.second );
}


class Test_union_different_iterator : public Test
{
    template <typename T>
    class WR1
    {
    public:
	    WR1() : _data( T() ) {}
	    WR1( T data ) : _data( data ) {}
	    const WR1<T> operator=( const WR1<T>& other )
	    {
		    _data = other._data;
		    return(*this);
	    }
	    bool operator==( const WR1<T>& other ) const
	    {
		    return _data == other._data;
	    }
	    T _data;
    };

    template <typename T>
    class WR2
    {
    public:
	    WR2() : _data( T() ) {}
	    WR2( T data ) : _data( data ) {}
	    const WR2<T> operator=( const WR2<T>& other )
	    {
		    _data = other._data;
		    return(*this);
	    }
	    bool operator==( const WR2<T>& other ) const
	    {
		    return _data == other._data;
	    }

	    T _data;
    };

public:
	void startTest()
	{
        using namespace std;

        // vInt: 0 1 2 3 4 5 6 7 8 9
		// vStr: A B C D E F G H I J
		typedef vector<Integer> Cont1;			// container 1
		typedef vector<String> Cont2;			// container 2
		typedef vector<Integer>::iterator IterCont1;
		typedef vector<String>::iterator IterCont2;

		// Test setup
		const int VECTOR_SIZE = 10;

		Cont1 vInt;
		Cont2 vStr;

		for ( int i = 0; i < VECTOR_SIZE; ++i )
		{
			vInt.push_back( i );
			vStr.push_back( string( 1, i + 65 ) );
		}

		print_vector( vInt );
		print_vector( vStr );

		// Test -1: constructor
		{
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui_end( vInt.end(), vInt.end(), vStr.end(), vStr.end() );
		}

		// Test 0: make_union_iterator
#if 0
		{
			print_union_different( make_union_different_iterator<IterCont1, IterCont2, WR1, WR2>
										( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() ),
								   make_union_different_iterator<IterCont1, IterCont2, WR1, WR2>
										( vInt.end(), vInt.end(), vStr.end(), vStr.end() ) );

			typedef pair< IterCont1, IterCont1 > Range1;
			typedef pair< IterCont2, IterCont2 > Range2;
			Range1 range_a( vInt.begin(), vInt.end() );
			Range2 range_b( vStr.begin(), vStr.end() );
			Range1 range_c( vInt.end(), vInt.end() );
			Range2 range_d( vStr.end(), vStr.end() );
			print_union_different( make_union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				                       ( range_a, range_b ),
								   make_union_different_iterator<IterCont1, IterCont2, WR1, WR2>
								       ( range_c, range_d ) );
		}

		// Test 0: make_union_range
		{
			print_union_different( make_union_different_range<IterCont1, IterCont2, WR1, WR2>
				                       ( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() ) );

			typedef pair< IterCont1, IterCont1 > Range1;
			typedef pair< IterCont2, IterCont2 > Range2;
			Range1 range_a( vInt.begin(), vInt.end() );
			Range2 range_b( vStr.begin(), vStr.end() );
			print_union_different( make_union_different_range<IterCont1, IterCont2, WR1, WR2>
				                       ( range_a, range_b ) );
		}
#endif

		// Test 1: operator==()
		{
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui1( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui2( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );

			assert( ui1 == ui2 );
		}

		// Test 2: operator!=()
		{
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui1( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui2( vInt.end(), vInt.end(), vStr.end(), vStr.end() );
			assert( ui1 != ui2 );
		}

		// Test 3: operator*() read-only, operator++(), operator++(int), operator--(), operator--(int)
		{
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui1( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );

			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 0 ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 1 ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 2 ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 3 ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 4 ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 5 ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 6 ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 7 ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 8 ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 9 ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "A" ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "B" ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "C" ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "D" ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "E" ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "F" ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "G" ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "H" ) ) ) ) == *ui1 ); ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "I" ) ) ) ) == *ui1 ); ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "J" ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "I" ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "H" ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "G" ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "F" ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "E" ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "D" ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "C" ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "B" ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR2<String>( String( "A" ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 9 ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 8 ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 7 ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 6 ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 5 ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 4 ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 3 ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 2 ) ) ) ) == *ui1 ); --ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 1 ) ) ) ) == *ui1 ); ui1--;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 0 ) ) ) ) == *ui1 );
		}

		// Test 5: copy constructor
		{
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui1( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );

			ui1++; ui1++; ui1++;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 3 ) ) ) ) == *ui1 );

			union_different_iterator<IterCont1, IterCont2, WR1, WR2> ui2 = ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 3 ) ) ) ) == *ui2 );

			union_different_iterator<IterCont1, IterCont2, WR1, WR2>ui3( ui1 );
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 3 ) ) ) ) == *ui3 );
		}

		// Test 6: default constructor, operator=()
		{
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui1( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );
			++ui1; ++ui1; ++ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 3 ) ) ) ) == *ui1 );

			union_different_iterator<IterCont1, IterCont2, WR1, WR2> ui2;
			ui2 = ui1;
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 3 ) ) ) ) == *ui2 );
		}

		// Test 7: verifica di scrivibilit� sull'oggetto tornato da operator*()
		{
			union_different_iterator<IterCont1, IterCont2, WR1, WR2>
				ui1( vInt.begin(), vInt.end(), vStr.begin(), vStr.end() );
            ++ui1; ++ui1; ++ui1;

            boost::variant< WR1<Integer>, WR2<String> > element = *ui1;
            element = WR1<Integer>( Integer( 30 ) );
			assert( ( boost::variant<WR1<Integer>, WR2<String> >( WR1<Integer>( Integer( 30 ) ) ) ) == element );
		}

		cout << "All test passed" << endl;

	}

};


#endif