
#ifndef _UNION_DIFFERENT_ITERATOR
#define _UNION_DIFFERENT_ITERATOR

/*

Note:
- � un forward iterator in sola lettura, quindi gli manca il concept output
  iterator; perci� non implementa operator*() in scrittura;
- il value_type � definito come
    boost::variant<
        WR_A<typename std::iterator_traits<FI1>::value_type>,
        WR_B<typename std::iterator_traits<FI2>::value_type>
    >
*/

#include <iterator>
#include <utility>

#include <boost/variant.hpp>

template <typename FI1, typename FI2, template <typename> class WR_A, template <typename> class WR_B>
class union_different_iterator
	: public std::iterator< std::forward_iterator_tag,
                            boost::variant<
                              WR_A<typename std::iterator_traits<FI1>::value_type>,
                              WR_B<typename std::iterator_traits<FI2>::value_type>
                            >
                          >
{
    typedef std::iterator<
        std::forward_iterator_tag,
        boost::variant<
            WR_A<typename std::iterator_traits<FI1>::value_type>,
            WR_B<typename std::iterator_traits<FI2>::value_type>
        >
    > Base;
public:
	typedef union_different_iterator<FI1, FI2, WR_A, WR_B> self;

	typedef typename std::iterator_traits<FI1>::value_type value_type1;
	typedef typename std::iterator_traits<FI2>::value_type value_type2;

	union_different_iterator()
	{}

	union_different_iterator( FI1 begin_a, FI1 end_a, FI2 begin_b, FI2 end_b )
		: _begin_a( begin_a ), _end_a( end_a ), _curr_a( begin_a ),
		  _begin_b( begin_b ), _end_b( end_b ), _curr_b( begin_b )
	{}

	union_different_iterator( const self& other )
		: _begin_a( other._begin_a ), _end_a( other._end_a ), _curr_a( other._curr_a ),
		  _begin_b( other._begin_b ), _end_b( other._end_b ), _curr_b( other._curr_b )
	{}

	const self& operator=( const self& other )
	{
		_begin_a = other._begin_a;
		_end_a = other._end_a;
		_curr_a = other._curr_a;

		_begin_b = other._begin_b;
		_end_b = other._end_b;
		_curr_b = other._curr_b;

		return(*this);
	}

	bool operator==( const self& other ) const
	{
		return ( _curr_a == other._curr_a ) && ( _curr_b == other._curr_b );
	}

	bool operator!=( const self& other ) const
	{
		return !this->operator == ( other );
	}

    // Normalmente operator*() restituisce un reference, in modo da poter
    // essere scritto oltre che letto; in questo caso per� l'utente di
    // union_different_iterator dovrebbe vedere un riferimento ad un
    // variant (vedi value_type), che per� non esiste in memoria; esistono
    // solo i reference agli elementi singoli dei due iterator uniti.
    // Il variant da restituire va quindi costruito qui e tornato per valore.
	typename Base::value_type operator*() const
	{
		return _curr_a == _end_a ?
			   typename Base::value_type( WR_B<value_type2>( *_curr_b ) ) :
               typename Base::value_type( WR_A<value_type1>( *_curr_a ) ) ;
	}

	self& operator++()
	{
		if ( _curr_a == _end_a )
			++_curr_b;
		else
			++_curr_a;
		return *this;
	}

	self operator++(int)
	{
		self tmp = *this; ++( *this ); return tmp;
	}

	self& operator--()
	{
		if (_curr_b == _begin_b )
			--_curr_a;
		else
			--_curr_b;
		return *this;
	}

	self operator--(int)
	{
		self tmp = *this; --( *this ); return tmp;
	}

	template <typename T>
	bool isTypeOf()
	{
		if ( _curr_a == _end_a )
			return boost::is_same<WR_B<T>, WR_B<value_type2> >::value;
		else
			return boost::is_same<WR_A<T>, WR_A<value_type1> >::value;
	}

private:
	FI1 _begin_a;
	FI1 _end_a;
	FI1 _curr_a;

	FI2 _begin_b;
	FI2 _end_b;
	FI2 _curr_b;
};

template <typename FI1, typename FI2, template <typename> class WR_A, template <typename> class WR_B>
inline union_different_iterator<FI1, FI2, WR_A, WR_B>
make_union_different_iterator(FI1 begin_a, FI1 end_a, FI2 begin_b, FI2 end_b) {
  return union_different_iterator<FI1, FI2, WR_A, WR_B>(begin_a, end_a, begin_b, end_b);
}

template <typename FI1, typename FI2, template <typename> class WR_A, template <typename> class WR_B>
inline union_different_iterator<FI1, FI2, WR_A, WR_B>
make_union_different_iterator( std::pair<FI1, FI1> range_a, std::pair<FI2, FI2> range_b ) {
  return union_different_iterator<FI1, FI2, WR_A, WR_B>(range_a.first, range_a.second,
		                                                range_b.first, range_b.second);
}

template <typename FI1, typename FI2, template <typename> class WR_A, template <typename> class WR_B>
inline std::pair< union_different_iterator<FI1, FI2, WR_A, WR_B>,
             union_different_iterator<FI1, FI2, WR_A, WR_B> >
make_union_different_range( FI1 begin_a, FI1 end_a, FI2 begin_b, FI2 end_b ) {
  return std::make_pair( make_union_different_iterator<FI1, FI2, WR_A, WR_B>( begin_a, end_a, begin_b, end_b ),
                         make_union_different_iterator<FI1, FI2, WR_A, WR_B>( end_a, end_a, end_b, end_b ) );
}

template <typename FI1, typename FI2, template <typename> class WR_A, template <typename> class WR_B>
inline std::pair< union_different_iterator<FI1, FI2, WR_A, WR_B>,
             union_different_iterator<FI1, FI2, WR_A, WR_B> >
make_union_different_range( std::pair<FI1, FI1> range_a, std::pair<FI2, FI2> range_b ) {
  return std::make_pair( make_union_different_iterator<FI1, FI2, WR_A, WR_B>( range_a, range_b ),
                         make_union_different_iterator<FI1, FI2, WR_A, WR_B>( range_a.second, range_a.second,
					                                                          range_b.second, range_b.second ) );
}




/*
 * NOTE: la differenza tra
 *
 * reference operator*() const
 *
 * e
 *
 * reference operator*()
 *
 * sta nel fatto che la prima risponde ad oggetti di tipo
 * const union_different_iterator<...>, mentre la seconda risponde
 * ad oggetti non-const
 *
 * Se invece si definisce un oggetto non-const di tipo
 * union_different_iterator<container<T>::const_iterator>, automaticamente
 * il tipo reference diventa un reference a const, quindi
 * viene utilizzata sempre la seconda forma di operator*()
 */

#endif